# file auto generated with scripts/tool_index_to_id_map.py
DISTRO_TOOLS_ID_TO_PATH = {
    "ucsc_table_direct1": "data_source/ucsc_tablebrowser.xml",
    "MAF_To_Fasta1": "maf/maf_to_fasta.xml",
    "ratmine": "data_source/ratmine.xml",
    "mousemine": "data_source/mousemine.xml",
    "__SORTLIST__": "${model_tools_path}/sort_collection_list.xml",
    "__UNZIP_COLLECTION__": "${model_tools_path}/unzip_collection.xml",
    "cbi_rice_mart": "data_source/cbi_rice_mart.xml",
    "ucsc_table_direct_archaea1": "data_source/ucsc_tablebrowser_archaea.xml",
    "wig_to_bigWig": "filters/wig_to_bigwig.xml",
    "ebi_sra_main": "data_source/ebi_sra.xml",
    "metabolicmine": "data_source/metabolicmine.xml",
    "Extract_features1": "filters/gff/extract_GFF_Features.xml",
    "upload1": "data_source/upload.xml",
    "wc_gnu": "filters/wc_gnu.xml",
    "join1": "filters/joiner.xml",
    "random_lines1": "filters/randomlines.xml",
    "modENCODEfly": "data_source/fly_modencode.xml",
    "gff_filter_by_attribute": "filters/gff/gff_filter_by_attribute.xml",
    "gtf2bedgraph": "filters/gtf2bedgraph.xml",
    "ChangeCase": "filters/changeCase.xml",
    "__FLATTEN__": "${model_tools_path}/flatten_collection.xml",
    "wiggle2simple1": "filters/wiggle_to_simple.xml",
    "GeneBed_Maf_Fasta2": "maf/genebed_maf_to_fasta.xml",
    "trimmer": "filters/trimmer.xml",
    "createInterval": "filters/CreateInterval.xml",
    "gff_filter_by_feature_count": "filters/gff/gff_filter_by_feature_count.xml",
    "Interval2Maf1": "maf/interval2maf.xml",
    "genomespace_exporter": "genomespace/genomespace_exporter.xml",
    "Show tail1": "filters/tailWrapper.xml",
    "barchart_gnuplot": "plotting/bar_chart.xml",
    "microbial_import1": "data_source/microbial_import.xml",
    "axt_to_concat_fasta": "filters/axt_to_concat_fasta.xml",
    "__FILTER_FROM_FILE__": "${model_tools_path}/filter_from_file.xml",
    "Interval2Maf_pairwise1": "maf/interval2maf_pairwise.xml",
    "Show beginning1": "filters/headWrapper.xml",
    "axt_to_lav_1": "filters/axt_to_lav.xml",
    "modmine": "data_source/modmine.xml",
    "__EXPORT_HISTORY__": "${model_tools_path}/imp_exp/exp_history_to_archive.xml",
    "MAF_Thread_For_Species1": "maf/maf_thread_for_species.xml",
    "vcf_to_maf_customtrack1": "maf/vcf_to_maf_customtrack.xml",
    "MAF_To_BED1": "maf/maf_to_bed.xml",
    "__DATA_FETCH__": "${model_tools_path}/data_fetch.xml",
    "__IMPORT_HISTORY__": "${model_tools_path}/imp_exp/imp_history_from_archive.xml",
    "biomart": "data_source/biomart.xml",
    "Sff_extractor": "filters/sff_extractor.xml",
    "secure_hash_message_digest": "filters/secure_hash_message_digest.xml",
    "MAF_Reverse_Complement_1": "maf/maf_reverse_complement.xml",
    "mergeCols1": "filters/mergeCols.xml",
    "gff2bed1": "filters/gff2bed.xml",
    "Grouping1": "stats/grouping.xml",
    "maf_limit_size1": "maf/maf_limit_size.xml",
    "sort1": "filters/sorter.xml",
    "Convert characters1": "filters/convert_characters.xml",
    "MAF_To_Interval1": "maf/maf_to_interval.xml",
    "MAF_filter": "maf/maf_filter.xml",
    "MAF_split_blocks_by_species1": "maf/maf_split_by_species.xml",
    "genomespace_importer": "genomespace/genomespace_importer.xml",
    "gene2exon1": "filters/ucsc_gene_bed_to_exon_bed.xml",
    "Cut1": "filters/cutWrapper.xml",
    "Count1": "filters/uniq.xml",
    "MAF_Limit_To_Species1": "maf/maf_limit_to_species.xml",
    "ucsc_table_direct_test1": "data_source/ucsc_tablebrowser_test.xml",
    "wormbase": "data_source/wormbase.xml",
    "maf_stats1": "maf/maf_stats.xml",
    "zebrafishmine": "data_source/zebrafishmine.xml",
    "Paste1": "filters/pasteWrapper.xml",
    "Interval_Maf_Merged_Fasta2": "maf/interval_maf_to_merged_fasta.xml",
    "modENCODEworm": "data_source/worm_modencode.xml",
    "gtf_filter_by_attribute_values_list": "filters/gff/gtf_filter_by_attribute_values_list.xml",
    "Summary_Statistics1": "stats/gsummary.xml",
    "qual_stats_boxplot": "plotting/boxplot.xml",
    "cat1": ["filters/catWrapper.xml", "filters/catWrapper2.xml"],
    "maf_by_block_number1": "maf/maf_by_block_number.xml",
    "Grep1": "filters/grep.xml",
    "eupathdb": "data_source/eupathdb.xml",
    "__RELABEL_FROM_FILE__": "${model_tools_path}/relabel_from_file.xml",
    "__ZIP_COLLECTION__": "${model_tools_path}/zip_collection.xml",
    "lav_to_bed1": "filters/lav_to_bed.xml",
    "comp1": "filters/compare.xml",
    "bed_to_bigBed": "filters/bed_to_bigbed.xml",
    "liftOver1": "extract/liftOver_wrapper.xml",
    "bed2gff1": "filters/bed2gff.xml",
    "__MERGE_COLLECTION__": "${model_tools_path}/merge_collection.xml",
    "gramenemart": "data_source/gramene_mart.xml",
    "hbvar": "data_source/hbvar.xml",
    "yeastmine": "data_source/yeastmine.xml",
    "Filter1": "stats/filtering.xml",
    "flymine": "data_source/flymine.xml",
    "wormbase_test": "data_source/wormbase_test.xml",
    "__FILTER_FAILED_DATASETS__": "${model_tools_path}/filter_failed_collection.xml",
    "Extract genomic DNA 1": "extract/extract_genomic_dna.xml",
    "Remove beginning1": "filters/remove_beginning.xml",
    "flymine_test": "data_source/flymine_test.xml",
    "axt_to_fasta": "filters/axt_to_fasta.xml",
    "addValue": "filters/fixedValueColumn.xml",
    "__TAG_FROM_FILE__": "${model_tools_path}/tag_collection_from_file.xml",
    "param_value_from_file": "expression_tools/parse_values_from_file.xml",
    "__APPLY_RULES__": "${model_tools_path}/apply_rules.xml",
}
