/* bootstrap custom style stuff - taken from demo template. */
/*
 * Base structure
 */

/* Move down content because we have a fixed navbar that is 50px tall */
body {
  padding-top: 50px;
}


/*
 * Global add-ons
 */

.sub-header {
  padding-bottom: 10px;
  border-bottom: 1px solid #eee;
}

/*
 * Top navigation
 * Hide default border to remove 1px line.
 */
.navbar-fixed-top {
  border: 0;
}

/*
 * Sidebar
 */

/* Hide for mobile, show later */
.sidebar {
  display: none;
}
@media (min-width: 768px) {
  .sidebar {
    position: fixed;
    top: 51px;
    bottom: 0;
    left: 0;
    z-index: 1000;
    display: block;
    padding: 20px;
    overflow-x: hidden;
    overflow-y: auto; /* Scrollable contents if viewport is shorter than content. */
    background-color: #f5f5f5;
    border-right: 1px solid #eee;
  }
}

/* Sidebar navigation */
.nav-sidebar {
  margin-right: -21px; /* 20px padding + 1px border */
  margin-bottom: 20px;
  margin-left: -20px;
}
.nav-sidebar > li > a {
  padding-right: 20px;
  padding-left: 20px;
}
.nav-sidebar > .active > a,
.nav-sidebar > .active > a:hover,
.nav-sidebar > .active > a:focus {
  color: #fff;
  background-color: #428bca;
}

/* Table in github markdown style (https://gist.github.com/andyferra/2554919#file-github-css-L154) */
table {
  padding: 0;
  margin-bottom: 10px;
}
table tr {
  border-top: 1px solid #cccccc;
  background-color: white;
  margin: 0;
  padding: 0;
}
table tr:nth-child(2n) {
  background-color: #f8f8f8;
}
table tr th {
  font-weight: bold;
  border: 1px solid #cccccc;
  text-align: left;
  margin: 0;
  padding: 6px 13px;
}
table tr td {
  border: 1px solid #cccccc;
  text-align: left;
  margin: 0;
  padding: 6px 13px;
}
table tr th :first-child, table tr td :first-child {
  margin-top: 0;
}
table tr th :last-child, table tr td :last-child {
  margin-bottom: 1;
}


/*
 * Main content
 */

.main {
  padding: 20px;
}
@media (min-width: 768px) {
  .main {
    padding-right: 40px;
    padding-left: 40px;
  }
}
.main .page-header {
  margin-top: 0;
}

summary {
  display: revert;
}

ul {
  list-style-type: none;
  padding-left: 10px;
}

li {
  padding: 2px;
}

pre {
  max-height: 500px;
  overflow: auto
}

.rcorners {
  border-radius: 10px;
  padding: 0px;
  border-style: solid;
}

.light-green {
  border: 2px transparent #73AD21;
  border-color: #dff0d8;
}

summary.light-green {
  background-color: #dff0d8
}

.light-red {
  /*  */
  border: 2px transparent #f2dede;
  border-color: #f2dede;
}

summary.light-red {
  background-color: #f2dede;
}

.padded {
  background-color: #fff;
  margin: 10px;
}

.tooltip .tooltiptext {
  visibility: hidden;
  width: 120px;
  background-color: black;
  color: #fff;
  text-align: center;
  border-radius: 6px;
  padding: 5px 0;

  /* Position the tooltip */
  position: absolute;
  z-index: 1;
}

.tooltip:hover .tooltiptext {
  visibility: visible;
}

/*
 * Placeholder dashboard ideas
 */

.placeholders {
  margin-bottom: 30px;
  text-align: center;
}
.placeholders h4 {
  margin-bottom: 0;
}
.placeholder {
  margin-bottom: 20px;
}
.placeholder img {
  display: inline-block;
  border-radius: 50%;
}

.text-success-custom{
    color: rgba(6, 108, 8, 1);
}
.text-danger-custom{
    font-weight: bold;
    color: rgba(213, 24, 20, 1);
}
.panel-danger-custom .panel-title{
    font-weight: bold;
    color: rgba(213, 24, 20, 1);
}t status

.panel-success-custom .panel-title{
    color: rgba(6, 108, 8, 1);
}t status
